@@ -12,7 +12,16 @@
 #include "zend_language_scanner_defs.h"
 #include "zend_language_parser.h"
 #include "zend_exceptions.h"
+#include "zend_hash.h"
 #include "zend_smart_str.h"
+#if PHP_VERSION_ID >= 80200
+/* Used for AllowDynamicProperties */
+#include "zend_attributes.h"
+#endif
+
+#ifndef ZEND_THIS
+#define ZEND_THIS getThis()
+#endif
 
 #ifndef ZEND_ARG_INFO_WITH_DEFAULT_VALUE
 #define ZEND_ARG_INFO_WITH_DEFAULT_VALUE(pass_by_ref, name, default_value) \
@@ -46,12 +55,25 @@
 #define ast_register_flag_constant(name, value) \
 	REGISTER_NS_LONG_CONSTANT("ast\\flags", name, value, CONST_CS | CONST_PERSISTENT)
 
-#define AST_CACHE_SLOT_KIND     &AST_G(cache_slots)[3 * 0]
-#define AST_CACHE_SLOT_FLAGS    &AST_G(cache_slots)[3 * 1]
-#define AST_CACHE_SLOT_LINENO   &AST_G(cache_slots)[3 * 2]
-#define AST_CACHE_SLOT_CHILDREN &AST_G(cache_slots)[3 * 3]
+// These are in the order the properties were declared with ast_declare_property.
+#define AST_NODE_PROP_KIND(object) OBJ_PROP_NUM((object), 0)
+#define AST_NODE_PROP_FLAGS(object) OBJ_PROP_NUM((object), 1)
+#define AST_NODE_PROP_LINENO(object) OBJ_PROP_NUM((object), 2)
+#define AST_NODE_PROP_CHILDREN(object) OBJ_PROP_NUM((object), 3)
 
-#define AST_CURRENT_VERSION 80
+#define AST_NODE_SET_PROP_KIND(object, num)       ZVAL_LONG(AST_NODE_PROP_KIND((object)), (num))
+#define AST_NODE_SET_PROP_FLAGS(object, num)      ZVAL_LONG(AST_NODE_PROP_FLAGS((object)), (num))
+#define AST_NODE_SET_PROP_LINENO(object, num)     ZVAL_LONG(AST_NODE_PROP_LINENO((object)), (num))
+// Set the ast\Node->children array to the given reference-counted array without incrementing the reference count of the array.
+// Do not use this macro with immutable arrays.
+#define AST_NODE_SET_PROP_CHILDREN_ARRAY(object, array) ZVAL_ARR(AST_NODE_PROP_CHILDREN((object)), (array))
+
+#define AST_METADATA_PROP_KIND(object) OBJ_PROP_NUM((object), 0)
+#define AST_METADATA_PROP_NAME(object) OBJ_PROP_NUM((object), 1)
+#define AST_METADATA_PROP_FLAGS(object) OBJ_PROP_NUM((object), 2)
+#define AST_METADATA_PROP_FLAGS_COMBINABLE(object) OBJ_PROP_NUM((object), 3)
+
+#define AST_CURRENT_VERSION 90
 
 /* Additional flags for BINARY_OP */
 #define AST_BINARY_IS_GREATER 256
@@ -98,6 +120,10 @@
 # define PARAM_MODIFIER_PRIVATE   ZEND_ACC_PRIVATE
 #endif
 
+#if PHP_VERSION_ID < 80100
+# define IS_NEVER 22
+#endif
+
 /* This contains state of the ast Node creator. */
 typedef struct ast_state_info {
 	zend_long version;
@@ -132,6 +158,7 @@ static const char *class_flags[] = {
 	AST_FLAG(CLASS_TRAIT),
 	AST_FLAG(CLASS_INTERFACE),
 	AST_FLAG(CLASS_ANONYMOUS),
+	AST_FLAG(CLASS_ENUM),
 	NULL
 };
 
@@ -158,6 +185,7 @@ static const char *type_flags[] = {
 	AST_FLAG(TYPE_ITERABLE),
 	AST_FLAG(TYPE_STATIC),
 	AST_FLAG(TYPE_MIXED),
+	AST_FLAG(TYPE_NEVER),
 	NULL
 };
 
@@ -261,7 +289,8 @@ static const char *closure_use_flags[] = {
 	AST_VISIBILITY_FLAGS, \
 	AST_FLAG(MODIFIER_STATIC), \
 	AST_FLAG(MODIFIER_ABSTRACT), \
-	AST_FLAG(MODIFIER_FINAL)
+	AST_FLAG(MODIFIER_FINAL), \
+	AST_FLAG(MODIFIER_READONLY)
 
 static const char *visibility_flags[] = {
 	AST_VISIBILITY_FLAGS,
@@ -294,7 +323,7 @@ static const char *conditional_flags[] = {
 
 static const ast_flag_info flag_info[] = {
 	{ AST_NAME, 0, name_flags },
-	{ ZEND_AST_CLASS, 0, class_flags },
+	{ ZEND_AST_CLASS, 1, class_flags },
 	{ ZEND_AST_PARAM, 1, param_flags },
 	{ ZEND_AST_TYPE, 0, type_flags },
 	{ ZEND_AST_CAST, 0, type_flags },
@@ -315,32 +344,41 @@ static const ast_flag_info flag_info[] = {
 	{ ZEND_AST_ARROW_FUNC, 1, func_flags },
 	{ ZEND_AST_PROP_DECL, 1, modifier_flags },
 	{ ZEND_AST_PROP_GROUP, 1, modifier_flags },
-	{ ZEND_AST_CLASS_CONST_DECL, 1, visibility_flags },
-	{ ZEND_AST_CLASS_CONST_GROUP, 1, visibility_flags },
+	{ ZEND_AST_CLASS_CONST_DECL, 1, modifier_flags },
+	{ ZEND_AST_CLASS_CONST_GROUP, 1, modifier_flags },
 	{ ZEND_AST_TRAIT_ALIAS, 1, modifier_flags },
 	{ ZEND_AST_DIM, 1, dim_flags },
 	{ ZEND_AST_CONDITIONAL, 1, conditional_flags },
 };
 
-// NOTE(tandre) in php 8, to get a writeable pointer to a property, OBJ_PROP_NUM(AST_CACHE_SLOT_PROPNAME) can be used.
-
-static inline void ast_update_property(zval *object, zend_string *name, zval *value, void **cache_slot) {
+static inline void ast_update_property(zval *object, zend_string *name, zval *value) {
 #if PHP_VERSION_ID < 80000
 	zval name_zv;
 	ZVAL_STR(&name_zv, name);
-	Z_OBJ_HT_P(object)->write_property(object, &name_zv, value, cache_slot);
+	Z_OBJ_HT_P(object)->write_property(object, &name_zv, value, NULL);
 #else
-	Z_OBJ_HT_P(object)->write_property(Z_OBJ_P(object), name, value, cache_slot);
+	Z_OBJ_HT_P(object)->write_property(Z_OBJ_P(object), name, value, NULL);
 #endif
 }
 
-static inline void ast_update_property_long(zval *object, zend_string *name, zend_long value_raw, void **cache_slot) {
+static inline void ast_update_property_long(zval *object, zend_string *name, zend_long value_raw) {
 	zval value_zv;
 	ZVAL_LONG(&value_zv, value_raw);
-	ast_update_property(object, name, &value_zv, cache_slot);
+	ast_update_property(object, name, &value_zv);
 }
 
-static zend_ast *get_ast(zend_string *code, zend_arena **ast_arena, char *filename) {
+static zend_ast *get_ast(zend_string *code, zend_arena **ast_arena, zend_string *filename) {
+#if PHP_VERSION_ID >= 80100
+	if (filename) {
+		return zend_compile_string_to_ast(code, ast_arena, filename);
+	} else {
+		zend_ast *ast;
+		filename = zend_string_init("string code", sizeof("string code") - 1, 0);
+		ast = zend_compile_string_to_ast(code, ast_arena, filename);
+		zend_string_release_ex(filename, 0);
+		return ast;
+	}
+#else
 	zval code_zv;
 	zend_bool original_in_compilation;
 	zend_lex_state original_lex_state;
@@ -352,7 +390,7 @@ static zend_ast *get_ast(zend_string *code, zend_arena **ast_arena, char *filena
 	CG(in_compilation) = 1;
 
 	zend_save_lexical_state(&original_lex_state);
-	zend_prepare_string_for_scanning(&code_zv, filename);
+	zend_prepare_string_for_scanning(&code_zv, filename ? filename->val : "string code");
 	CG(ast) = NULL;
 	CG(ast_arena) = zend_arena_create(1024 * 32);
 	LANG_SCNG(yy_state) = yycINITIAL;
@@ -373,6 +411,7 @@ static zend_ast *get_ast(zend_string *code, zend_arena **ast_arena, char *filena
 	zval_dtor(&code_zv);
 
 	return ast;
+#endif
 }
 
 /* Returns whether node->attr (i.e. flags) is used by this node kind. Not to be confused with php 8.0's attributes. */
@@ -405,6 +444,11 @@ static inline zend_bool ast_is_name(zend_ast *ast, zend_ast *parent, uint32_t i)
 	if (parent->kind == ZEND_AST_NAME_LIST) {
 		return 1;
 	}
+#if PHP_VERSION_ID >= 80100
+	if (parent->kind == ZEND_AST_TYPE_INTERSECTION) {
+		return 1;
+	}
+#endif
 #if PHP_VERSION_ID >= 80000
 	if (parent->kind == ZEND_AST_TYPE_UNION) {
 		return 1;
@@ -436,11 +480,22 @@ static inline zend_bool ast_is_name(zend_ast *ast, zend_ast *parent, uint32_t i)
 			|| parent->kind == ZEND_AST_METHOD;
 	}
 
+#if PHP_VERSION_ID >= 80100
+	if (i == 4) {
+		return parent->kind == ZEND_AST_CLASS;
+	}
+#endif
+
 	return 0;
 }
 
 /* Assumes that ast_is_name is already true */
 static inline zend_bool ast_is_type(zend_ast *ast, zend_ast *parent, uint32_t i) {
+#if PHP_VERSION_ID >= 80100
+	if (parent->kind == ZEND_AST_TYPE_INTERSECTION) {
+		return 1;
+	}
+#endif
 #if PHP_VERSION_ID >= 80000
 	if (parent->kind == ZEND_AST_TYPE_UNION) {
 		return 1;
@@ -460,6 +515,11 @@ static inline zend_bool ast_is_type(zend_ast *ast, zend_ast *parent, uint32_t i)
 #endif
 			|| parent->kind == ZEND_AST_METHOD;
 	}
+#if PHP_VERSION_ID >= 80100
+	if (i == 4) {
+		return parent->kind == ZEND_AST_CLASS;
+	}
+#endif
 	return 0;
 }
 
@@ -511,6 +571,7 @@ static const builtin_type_info builtin_types[] = {
 	{ZEND_STRL("false"), IS_FALSE},
 	// {ZEND_STRL("static"), IS_STATIC},  /* Impossible to be parsed before php 8 */
 	{ZEND_STRL("mixed"), IS_MIXED},
+	{ZEND_STRL("never"), IS_NEVER},
 	{NULL, 0, IS_UNDEF}
 };
 static inline zend_uchar lookup_builtin_type(const zend_string *name) {
@@ -545,10 +606,12 @@ static inline zend_ast_attr ast_assign_op_to_binary_op(zend_ast_attr attr) {
 }
 #endif
 
-static inline zend_ast **ast_get_children(zend_ast *ast, uint32_t *count) {
+static inline zend_ast **ast_get_children(zend_ast *ast, ast_state_info_t *state, uint32_t *count) {
 	if (ast_kind_is_decl(ast->kind)) {
 		zend_ast_decl *decl = (zend_ast_decl *) ast;
-#if PHP_VERSION_ID >= 80000
+#if PHP_VERSION_ID >= 80100
+		*count = decl->kind == ZEND_AST_CLASS ? (state->version >= 85 ? 5 : 4) : 5;
+#elif PHP_VERSION_ID >= 80000
 		*count = decl->kind == ZEND_AST_CLASS ? 4 : 5;
 #else
 		*count = decl->kind == ZEND_AST_CLASS ? 3 : 4;
@@ -569,30 +632,28 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state);
 static void ast_create_virtual_node_ex(
 		zval *zv, zend_ast_kind kind, zend_ast_attr attr, uint32_t lineno,
 		ast_state_info_t *state, uint32_t num_children, ...) {
-	zval tmp_zv;
 	va_list va;
 	uint32_t i;
 
 	object_init_ex(zv, ast_node_ce);
 
-	ast_update_property_long(zv, AST_STR(str_kind), kind, AST_CACHE_SLOT_KIND);
+	zend_object *obj = Z_OBJ_P(zv);
 
-	ast_update_property_long(zv, AST_STR(str_flags), attr, AST_CACHE_SLOT_FLAGS);
+	AST_NODE_SET_PROP_KIND(obj, kind);
+	AST_NODE_SET_PROP_FLAGS(obj, attr);
+	AST_NODE_SET_PROP_LINENO(obj, lineno);
 
-	ast_update_property_long(zv, AST_STR(str_lineno), lineno, AST_CACHE_SLOT_LINENO);
-
-	array_init(&tmp_zv);
-	Z_DELREF(tmp_zv);
-	ast_update_property(zv, AST_STR(str_children), &tmp_zv, AST_CACHE_SLOT_CHILDREN);
+	array_init_size(AST_NODE_PROP_CHILDREN(obj), num_children);
+	HashTable *children = Z_ARRVAL_P(AST_NODE_PROP_CHILDREN(obj));
 
 	va_start(va, num_children);
 	for (i = 0; i < num_children; i++) {
 		zval *child_zv = va_arg(va, zval *);
 		zend_string *child_name = ast_kind_child_name(kind, i);
 		if (child_name) {
-			zend_hash_add_new(Z_ARRVAL(tmp_zv), child_name, child_zv);
+			zend_hash_add_new(children, child_name, child_zv);
 		} else {
-			zend_hash_next_index_insert(Z_ARRVAL(tmp_zv), child_zv);
+			zend_hash_next_index_insert(children, child_zv);
 		}
 	}
 	va_end(va);
@@ -650,7 +711,7 @@ static inline void ast_name_to_zval(zend_ast *child, zend_ast *ast, zval *child_
 static void ast_fill_children_ht(HashTable *ht, zend_ast *ast, ast_state_info_t *state) {
 	uint32_t i, count;
 	zend_bool is_list = zend_ast_is_list(ast);
-	zend_ast **children = ast_get_children(ast, &count);
+	zend_ast **children = ast_get_children(ast, state, &count);
 	const zend_ast_kind ast_kind = ast->kind;
 	for (i = 0; i < count; ++i) {
 		zend_ast *child = children[i];
@@ -685,7 +746,7 @@ static void ast_fill_children_ht(HashTable *ht, zend_ast *ast, ast_state_info_t
 					}
 					break;
 				case ZEND_AST_CLASS:
-					if (i == 3) {
+					if (i >= 3) {
 						continue;
 					}
 					break;
@@ -696,6 +757,13 @@ static void ast_fill_children_ht(HashTable *ht, zend_ast *ast, ast_state_info_t
 					break;
 			}
 		}
+#if PHP_VERSION_ID >= 80100
+		if (ast_kind == ZEND_AST_CLASS && i == 4) {
+			if (state->version < 85)  {
+				continue;
+			}
+		}
+#endif
 #endif
 		/* This AST_CATCH check should occur before ast_is_name() */
 #if PHP_VERSION_ID < 70100
@@ -797,6 +865,17 @@ static void ast_fill_children_ht(HashTable *ht, zend_ast *ast, ast_state_info_t
 			zend_hash_add_new(ht, AST_STR(str_attributes), &tmp);
 		}
 #endif
+
+#if PHP_VERSION_ID < 80100
+		if (ast_kind == ZEND_AST_CLASS) {
+			if (state->version >= 85) {
+				zval tmp;
+				ZVAL_NULL(&tmp);
+				zend_hash_add_new(ht, AST_STR(str_type), &tmp);
+			}
+		}
+#endif
+
 		ZVAL_LONG(&id_zval, state->declIdCounter);
 		state->declIdCounter++;
 		zend_hash_add_new(ht, AST_STR(str___declId), &id_zval);
@@ -804,7 +883,7 @@ static void ast_fill_children_ht(HashTable *ht, zend_ast *ast, ast_state_info_t
 }
 
 static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
-	zval tmp_zv, children_zv;
+	zval tmp_zv;
 
 	if (ast == NULL) {
 		ZVAL_NULL(zv);
@@ -863,6 +942,7 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
 				ast_to_zval(zv, ast->child[0], state);
 				return;
 			}
+			break;
 #endif
 #if PHP_VERSION_ID >= 70400
 		case ZEND_AST_PROP_GROUP:
@@ -872,7 +952,7 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
 				ast_to_zval(zv, ast->child[1], state);
 				// The property visibility is on the AST_PROP_GROUP node.
 				// Add it to the AST_PROP_DECL node for old
-				ast_update_property_long(zv, AST_STR(str_flags), ast->attr, AST_CACHE_SLOT_FLAGS);
+				AST_NODE_SET_PROP_FLAGS(Z_OBJ_P(zv), ast->attr);
 				return;
 			}
 			break;
@@ -940,48 +1020,50 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
 
 	object_init_ex(zv, ast_node_ce);
 
-	ast_update_property_long(zv, AST_STR(str_kind), ast->kind, AST_CACHE_SLOT_KIND);
+	zend_object *obj = Z_OBJ_P(zv);
 
-	ast_update_property_long(zv, AST_STR(str_lineno), zend_ast_get_lineno(ast), AST_CACHE_SLOT_LINENO);
+	AST_NODE_SET_PROP_KIND(obj, ast->kind);
 
-	array_init(&children_zv);
-	Z_DELREF(children_zv);
-	ast_update_property(zv, AST_STR(str_children), &children_zv, AST_CACHE_SLOT_CHILDREN);
+	AST_NODE_SET_PROP_LINENO(obj, zend_ast_get_lineno(ast));
+
+	array_init(AST_NODE_PROP_CHILDREN(obj));
+	HashTable *children = Z_ARRVAL_P(AST_NODE_PROP_CHILDREN(obj));
 
 	if (ast_kind_is_decl(ast->kind)) {
 		zend_ast_decl *decl = (zend_ast_decl *) ast;
 
-		ast_update_property_long(zv, AST_STR(str_flags), decl->flags, AST_CACHE_SLOT_FLAGS);
+		AST_NODE_SET_PROP_FLAGS(obj, decl->flags);
 
-		ast_update_property_long(zv, AST_STR(str_endLineno), decl->end_lineno, NULL);
+		// This is an undeclared dynamic property and has no cache slot.
+		ast_update_property_long(zv, AST_STR(str_endLineno), decl->end_lineno);
 
 		if (decl->name) {
 			ZVAL_STR(&tmp_zv, decl->name);
+			Z_TRY_ADDREF(tmp_zv);
 		} else {
 			ZVAL_NULL(&tmp_zv);
 		}
 
-		Z_TRY_ADDREF(tmp_zv);
-		zend_hash_add_new(Z_ARRVAL(children_zv), AST_STR(str_name), &tmp_zv);
+		zend_hash_add_new(children, AST_STR(str_name), &tmp_zv);
 
 		if (decl->doc_comment) {
 			ZVAL_STR(&tmp_zv, decl->doc_comment);
+			Z_TRY_ADDREF(tmp_zv);
 		} else {
 			ZVAL_NULL(&tmp_zv);
 		}
 
-		Z_TRY_ADDREF(tmp_zv);
-		zend_hash_add_new(Z_ARRVAL(children_zv), AST_STR(str_docComment), &tmp_zv);
+		zend_hash_add_new(children, AST_STR(str_docComment), &tmp_zv);
 	} else {
 #if PHP_VERSION_ID < 70100
 		if (ast->kind == ZEND_AST_CLASS_CONST_DECL) {
 			ast->attr = ZEND_ACC_PUBLIC;
 		}
 #endif
-		ast_update_property_long(zv, AST_STR(str_flags), ast->attr, AST_CACHE_SLOT_FLAGS);
+		AST_NODE_SET_PROP_FLAGS(obj, ast->attr);
 	}
 
-	ast_fill_children_ht(Z_ARRVAL(children_zv), ast, state);
+	ast_fill_children_ht(children, ast, state);
 #if PHP_VERSION_ID < 70400
 	if (ast->kind == ZEND_AST_PROP_DECL && state->version >= 70) {
 		zval type_zval;
@@ -997,7 +1079,7 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
 			ast_create_virtual_node_ex(
 					zv, ZEND_AST_PROP_GROUP, ast->attr, zend_ast_get_lineno(ast), state, 2, &type_zval, &prop_group_zval);
 		}
-		ast_update_property_long(&prop_group_zval, AST_STR(str_flags), 0, AST_CACHE_SLOT_FLAGS);
+		AST_NODE_SET_PROP_FLAGS(obj, 0);
 	}
 #endif
 #if PHP_VERSION_ID < 80000
@@ -1006,7 +1088,7 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
 		zval attributes_zval;
 		ZVAL_COPY_VALUE(&const_decl_zval, zv);
 		ZVAL_NULL(&attributes_zval);
-		ast_update_property_long(zv, AST_STR(str_flags), 0, AST_CACHE_SLOT_FLAGS);
+		AST_NODE_SET_PROP_FLAGS(obj, 0);
 		// For version 80, create an AST_CLASS_CONST_GROUP wrapping the created AST_CLASS_CONST_DECL
 		ast_create_virtual_node_ex(
 				zv, ZEND_AST_CLASS_CONST_GROUP, ast->attr, zend_ast_get_lineno(ast), state, 2, &const_decl_zval, &attributes_zval);
@@ -1014,7 +1096,7 @@ static void ast_to_zval(zval *zv, zend_ast *ast, ast_state_info_t *state) {
 #endif
 }
 
-static const zend_long versions[] = {50, 60, 70, 80};
+static const zend_long versions[] = {50, 60, 70, 80, 85, 90};
 static const size_t versions_count = sizeof(versions)/sizeof(versions[0]);
 
 static inline zend_bool ast_version_deprecated(zend_long version) {
@@ -1105,7 +1187,7 @@ PHP_FUNCTION(parse_file) {
 		code = ZSTR_EMPTY_ALLOC();
 	}
 
-	ast = get_ast(code, &arena, filename->val);
+	ast = get_ast(code, &arena, filename);
 	if (!ast) {
 		zend_string_free(code);
 		return;
@@ -1135,7 +1217,7 @@ PHP_FUNCTION(parse_code) {
 		return;
 	}
 
-	ast = get_ast(code, &arena, filename ? filename->val : "string code");
+	ast = get_ast(code, &arena, filename);
 	if (!ast) {
 		return;
 	}
@@ -1188,21 +1270,21 @@ static inline const ast_flag_info *ast_get_flag_info(uint16_t ast_kind) {
 
 static void ast_build_metadata(zval *result) {
 	size_t i;
-	array_init(result);
+	array_init_size(result, ast_kinds_count);
 	for (i = 0; i < ast_kinds_count; i++) {
 		zend_ast_kind kind = ast_kinds[i];
 		const ast_flag_info *flag_info = ast_get_flag_info(kind);
 		zval info_zv, tmp_zv;
+		zend_object *obj;
 
 		object_init_ex(&info_zv, ast_metadata_ce);
+		obj = Z_OBJ(info_zv);
 
 		/* kind */
-		ast_update_property_long(&info_zv, AST_STR(str_kind), kind, NULL);
+		ZVAL_LONG(AST_METADATA_PROP_KIND(obj), kind);
 
 		/* name */
-		ZVAL_STRING(&tmp_zv, ast_kind_to_name(kind));
-		Z_TRY_DELREF(tmp_zv);
-		ast_update_property(&info_zv, AST_STR(str_name), &tmp_zv, NULL);
+		ZVAL_STRING(AST_METADATA_PROP_NAME(obj), ast_kind_to_name(kind));
 
 		/* flags */
 		array_init(&tmp_zv);
@@ -1212,12 +1294,10 @@ static void ast_build_metadata(zval *result) {
 				add_next_index_string(&tmp_zv, *flag);
 			}
 		}
-		Z_TRY_DELREF(tmp_zv);
-		ast_update_property(&info_zv, AST_STR(str_flags), &tmp_zv, NULL);
+		ZVAL_ARR(AST_METADATA_PROP_FLAGS(obj), Z_ARRVAL(tmp_zv));
 
 		/* flagsCombinable */
-		ZVAL_BOOL(&tmp_zv, flag_info && flag_info->combinable);
-		ast_update_property(&info_zv, AST_STR(str_flagsCombinable), &tmp_zv, NULL);
+		ZVAL_BOOL(AST_METADATA_PROP_FLAGS_COMBINABLE(obj), flag_info && flag_info->combinable);
 
 		add_index_zval(result, kind, &info_zv);
 	}
@@ -1273,49 +1353,34 @@ PHP_METHOD(ast_Node, __construct) {
 		Z_PARAM_LONG_EX(lineno, linenoNull, 1, 0)
 	ZEND_PARSE_PARAMETERS_END();
 
-	zval *zv = getThis();
+	zend_object *obj = Z_OBJ_P(ZEND_THIS);
 
 	switch (num_args) {
 		case 4:
 			if (!linenoNull) {
-				ast_update_property_long(zv, AST_STR(str_lineno), lineno, AST_CACHE_SLOT_LINENO);
+				AST_NODE_SET_PROP_LINENO(obj, lineno);
 			}
-			/* break missing intentionally */
+			/* Falls through - break missing intentionally */
 		case 3:
 			if (children != NULL) {
-				ast_update_property(zv, AST_STR(str_children), children, AST_CACHE_SLOT_CHILDREN);
+				ZVAL_COPY(AST_NODE_PROP_CHILDREN(obj), children);
 			}
-			/* break missing intentionally */
+			/* Falls through - break missing intentionally */
 		case 2:
 			if (!flagsNull) {
-				ast_update_property_long(zv, AST_STR(str_flags), flags, AST_CACHE_SLOT_FLAGS);
+				AST_NODE_SET_PROP_FLAGS(obj, flags);
 			}
-			/* break missing intentionally */
+			/* Falls through - break missing intentionally */
 		case 1:
 			if (!kindNull) {
-				ast_update_property_long(zv, AST_STR(str_kind), kind, AST_CACHE_SLOT_KIND);
+				AST_NODE_SET_PROP_KIND(obj, kind);
 			}
-			/* break missing intentionally */
+			/* Falls through - break missing intentionally */
 		case 0:
 			break;
 	}
 }
 
-const zend_function_entry ast_functions[] = {
-	ZEND_NS_FE("ast", parse_file, arginfo_parse_file)
-	ZEND_NS_FE("ast", parse_code, arginfo_parse_code)
-	ZEND_NS_FE("ast", get_kind_name, arginfo_get_kind_name)
-	ZEND_NS_FE("ast", kind_uses_flags, arginfo_kind_uses_flags)
-	ZEND_NS_FE("ast", get_metadata, arginfo_get_metadata)
-	ZEND_NS_FE("ast", get_supported_versions, arginfo_get_supported_versions)
-	PHP_FE_END
-};
-
-const zend_function_entry ast_node_functions[] = {
-	PHP_ME(ast_Node, __construct, arginfo_class_Node___construct, ZEND_ACC_PUBLIC)
-	PHP_FE_END
-};
-
 PHP_MINFO_FUNCTION(ast) {
 	zend_string *info = ast_version_info();
 
@@ -1329,7 +1394,6 @@ PHP_MINFO_FUNCTION(ast) {
 }
 
 PHP_RINIT_FUNCTION(ast) {
-	memset(AST_G(cache_slots), 0, sizeof(void *) * AST_NUM_CACHE_SLOTS);
 	ZVAL_UNDEF(&AST_G(metadata));
 	return SUCCESS;
 }
@@ -1362,6 +1426,7 @@ PHP_MINIT_FUNCTION(ast) {
 	ast_register_flag_constant("MODIFIER_STATIC", ZEND_ACC_STATIC);
 	ast_register_flag_constant("MODIFIER_ABSTRACT", ZEND_ACC_ABSTRACT);
 	ast_register_flag_constant("MODIFIER_FINAL", ZEND_ACC_FINAL);
+	ast_register_flag_constant("MODIFIER_READONLY", ZEND_ACC_READONLY);
 
 	ast_register_flag_constant("PARAM_MODIFIER_PUBLIC", PARAM_MODIFIER_PUBLIC);
 	ast_register_flag_constant("PARAM_MODIFIER_PROTECTED", PARAM_MODIFIER_PROTECTED);
@@ -1379,6 +1444,7 @@ PHP_MINIT_FUNCTION(ast) {
 	ast_register_flag_constant("CLASS_TRAIT", ZEND_ACC_TRAIT);
 	ast_register_flag_constant("CLASS_INTERFACE", ZEND_ACC_INTERFACE);
 	ast_register_flag_constant("CLASS_ANONYMOUS", ZEND_ACC_ANON_CLASS);
+	ast_register_flag_constant("CLASS_ENUM", ZEND_ACC_ENUM);
 
 	ast_register_flag_constant("PARAM_REF", ZEND_PARAM_REF);
 	ast_register_flag_constant("PARAM_VARIADIC", ZEND_PARAM_VARIADIC);
@@ -1396,6 +1462,7 @@ PHP_MINIT_FUNCTION(ast) {
 	ast_register_flag_constant("TYPE_ITERABLE", IS_ITERABLE);
 	ast_register_flag_constant("TYPE_STATIC", IS_STATIC);
 	ast_register_flag_constant("TYPE_MIXED", IS_MIXED);
+	ast_register_flag_constant("TYPE_NEVER", IS_NEVER);
 
 	ast_register_flag_constant("UNARY_BOOL_NOT", ZEND_BOOL_NOT);
 	ast_register_flag_constant("UNARY_BITWISE_NOT", ZEND_BW_NOT);
@@ -1462,12 +1529,16 @@ PHP_MINIT_FUNCTION(ast) {
 
 	ast_register_flag_constant("PARENTHESIZED_CONDITIONAL", ZEND_PARENTHESIZED_CONDITIONAL);
 
-	INIT_CLASS_ENTRY(tmp_ce, "ast\\Node", ast_node_functions);
+	INIT_CLASS_ENTRY(tmp_ce, "ast\\Node", class_ast_Node_methods);
 	ast_node_ce = zend_register_internal_class(&tmp_ce);
 	ast_declare_property(ast_node_ce, AST_STR(str_kind), &zv_null);
 	ast_declare_property(ast_node_ce, AST_STR(str_flags), &zv_null);
 	ast_declare_property(ast_node_ce, AST_STR(str_lineno), &zv_null);
 	ast_declare_property(ast_node_ce, AST_STR(str_children), &zv_null);
+#ifdef ZEND_ACC_ALLOW_DYNAMIC_PROPERTIES
+	zend_add_class_attribute(ast_node_ce, zend_ce_allow_dynamic_properties->name, 0);
+	ast_node_ce->ce_flags |= ZEND_ACC_ALLOW_DYNAMIC_PROPERTIES;
+#endif
 
 	INIT_CLASS_ENTRY(tmp_ce, "ast\\Metadata", NULL);
 	ast_metadata_ce = zend_register_internal_class(&tmp_ce);
@@ -1490,7 +1561,7 @@ PHP_MSHUTDOWN_FUNCTION(ast) {
 zend_module_entry ast_module_entry = {
 	STANDARD_MODULE_HEADER,
 	"ast",
-	ast_functions,
+	ext_functions,
 	PHP_MINIT(ast),
 	PHP_MSHUTDOWN(ast),
 	PHP_RINIT(ast),
