@@ -7,7 +7,7 @@
 extern zend_module_entry ast_module_entry;
 #define phpext_ast_ptr &ast_module_entry
 
-#define PHP_AST_VERSION "1.0.10"
+#define PHP_AST_VERSION "1.0.16"
 
 #ifdef PHP_WIN32
 #	define PHP_AST_API __declspec(dllexport)
@@ -21,12 +21,7 @@ extern zend_module_entry ast_module_entry;
 #include "TSRM.h"
 #endif
 
-// PHP 7.4 added a 3rd cache slot for property_info
-// and expects cache_slot[2] to be null.
-#define AST_NUM_CACHE_SLOTS (3 * 4)
-
 ZEND_BEGIN_MODULE_GLOBALS(ast)
-	void *cache_slots[AST_NUM_CACHE_SLOTS];
 	zval metadata;
 ZEND_END_MODULE_GLOBALS(ast)
 
@@ -75,6 +70,17 @@ extern ast_str_globals str_globals;
 // NOTE: The first hex digit is the number of child nodes a given kind has
 #endif
 
+#if PHP_VERSION_ID < 80100
+# define ZEND_ACC_ENUM (1 << 22)
+# define ZEND_ACC_READONLY (1 << 7)
+
+/* 0 child nodes */
+# define ZEND_AST_CALLABLE_CONVERT 3
+/* 3 child nodes - name, expr, attributes */
+# define ZEND_AST_ENUM_CASE 0x3fe
+# define ZEND_AST_TYPE_INTERSECTION ((1 << (ZEND_AST_IS_LIST_SHIFT + 1)) - 6)
+#endif
+
 /* Pretend it still exists */
 #if PHP_VERSION_ID >= 70100
 # define ZEND_AST_LIST ((1 << (ZEND_AST_IS_LIST_SHIFT + 1)) - 1)
